import zodiax
import jax.numpy as np
import jax.scipy as jsp
from jax import hessian, lax, Array
from typing import Union, List, Any


__all__ = ["poiss_loglike", "chi2_loglike", "covaraince_entropy", 
    "fisher_matrix", "covariance_matrix", "self_fisher_matrix", 
    "self_covariance_matrix"]


"""
Assumes all pytrees have a .model() function
TODO: Allow *args and **kwargs to be passed to pytree.model()
    -> Actually kind of hard, since we already have * args and ** kwargs for
    the loglike function
"""


Base = lambda: zodiax.base.Base
Params = Union[str, List[str]]


def poiss_loglike(pytree : Base(), data : Array) -> float:
    """
    Poissonian log likelihood of the pytree given the data. Assumes the pytree
    has a .model() function.

    Parameters
    ----------
    pytree : Base
        Pytree with a .model() function.
    data : Array
        Data to compare the model to.
    
    Returns
    -------
    log_likelihood : Array
        Log likelihood of the pytree given the data.
    """
    return jsp.stats.poisson.logpmf(pytree.model(), data).sum()


def chi2_loglike(pytree : Base(), data : Array, noise : float = 1) -> float:
    """
    Chi2 log likelihood of the pytree given the data. Assumes the pytree has a
    .model() function.

    Parameters
    ----------
    pytree : Base
        Pytree with a .model() function.
    data : Array
        Data to compare the model to.
    noise : float = 1
        The 'scale' parameter representing the noise in the data.
    
    Returns
    -------
    log_likelihood : Array
        Log likelihood of the pytree given the data.
    """
    return jsp.stats.chi2.logpdf(pytree.model(), data, scale=noise).sum()


def covaraince_entropy(covariance_matrix : Array) -> Array:
    """
    Calculates the entropy of a covariance matrix.

    Parameters
    ----------
    covariance_matrix : Array
        The covariance matrix to calculate the entropy of.
    
    Returns
    -------
    entropy : Array
        The entropy of the covariance matrix.
    """
    sign, logdet = np.linalg.slogdet(covariance_matrix)
    return 0.5 * (np.log(2 * np.pi * np.e) + (sign * logdet))


def fisher_matrix(
        pytree           : Base(), 
        parameters       : Params,
        loglike_fn       : callable,
        *loglike_args    : Any,
        shape_dict       : dict = {},
        **loglike_kwargs : Any,
        ) -> Array:
    """
    Calculates the Fisher information matrix of the pytree parameters. The
    `shaped_dict` parameter is used to specify the shape of the differentiated
    vector for specific parameters. For example, if the parameter `param` is
    a 1D array of shape (5,) and we wanted to calculate the Fisher information
    of the mean, we can pass in `shape_dict={'param': (1,)}`. This will
    differentiate the log likelihood with respect to the mean of the parameters.

    Parameters
    ----------
    pytree : Base
        Pytree with a .model() function.
    parameters : Union[str, list]
        A path or list of paths or list of nested paths.
    loglike_fn : callable
        The log likelihood function to differentiate.
    *loglike_args : Any
        The args to pass to the log likelihood function.
    shape_dict : dict = {}
        A dictionary specifying the shape of the differentiated vector for
        specific parameters.
    **loglike_kwargs : Any
        The kwargs to pass to the log likelihood function.
    
    Returns
    -------
    fisher_matrix : Array
        The Fisher information matrix of the pytree parameters.
    """
    # Build X vec
    pytree = zodiax.tree.set_array(pytree, parameters)
    shapes, lengths = _shapes_and_lengths(pytree, parameters, shape_dict)
    X = np.zeros(_lengths_to_N(lengths))

    # Build function to calculate FIM and calculate
    @hessian
    def calc_fim(X):
        parametric_pytree = _perturb(X, pytree, parameters, shapes, lengths)
        return loglike_fn(parametric_pytree, *loglike_args, **loglike_kwargs)
    return calc_fim(X)


def covariance_matrix(
        pytree           : Base(), 
        parameters       : Params,
        loglike_fn       : callable,
        *loglike_args    : Any,
        shape_dict       : dict = {},
        **loglike_kwargs : Any
        ) -> Array:
    """
    Calculates the covariance matrix of the pytree parameters. The
    `shaped_dict` parameter is used to specify the shape of the differentiated
    vector for specific parameters. For example, if the parameter `param` is
    a 1D array of shape (5,) and we wanted to calculate the covariance matrix
    of the mean, we can pass in `shape_dict={'param': (1,)}`. This will
    differentiate the log likelihood with respect to the mean of the parameters.

    Parameters
    ----------
    pytree : Base
        Pytree with a .model() function.
    parameters : Union[str, list]
        A path or list of paths or list of nested paths.
    loglike_fn : callable
        The log likelihood function to differentiate.
    *loglike_args : Any
        The args to pass to the log likelihood function.
    shape_dict : dict = {}
        A dictionary specifying the shape of the differentiated vector for
        specific parameters.
    **loglike_kwargs : Any
        The kwargs to pass to the log likelihood function.
    
    Returns
    -------
    covariance_matrix : Array
        The covariance matrix of the pytree parameters.
    """
    return -np.linalg.inv(fisher_matrix(pytree, parameters, loglike_fn, 
        *loglike_args, shape_dict=shape_dict, **loglike_kwargs))


def self_fisher_matrix(
        pytree           : Base(), 
        parameters       : Params, 
        loglike_fn       : callable,
        *loglike_args    : Any,
        shape_dict       : dict = {},
        **loglike_kwargs : Any
        ) -> Array:
    """
    Calculates the Fisher information matrix of the pytree parameters with
    respect to itself. This can be used to optimise the fisher information of
    the pytree model. Simply calcluates the data from the `.model()` method and
    passes it to the `fisher_matrix` function. The `shaped_dict` parameter is
    used to specify the shape of the differentiated vector for specific
    parameters. For example, if the parameter `param` is a 1D array of shape
    (5,) and we wanted to calculate the covariance matrix of the mean, we can
    pass in `shape_dict={'param': (1,)}`. This will differentiate the log
    likelihood with respect to the mean of the parameters.

    Parameters
    ----------
    pytree : Base
        Pytree with a .model() function.
    parameters : Union[str, list]
        A path or list of paths or list of nested paths.
    loglike_fn : callable
        The log likelihood function to differentiate.
    *loglike_args : Any
        The args to pass to the log likelihood function.
    shape_dict : dict = {}
        A dictionary specifying the shape of the differentiated vector for
        specific parameters.
    **loglike_kwargs : Any
        The kwargs to pass to the log likelihood function.
    
    Returns
    -------
    fisher_matrix : Array
        The Fisher information matrix of the pytree parameters.
    """
    # Build X vec and get data
    data = pytree.model()
    loglike_args = [pytree.model()] + list(loglike_args)
    return fisher_matrix(pytree, parameters, loglike_fn, *loglike_args, 
        shape_dict=shape_dict, **loglike_kwargs)


def self_covariance_matrix(
        pytree           : Base(), 
        parameters       : Params, 
        loglike_fn       : callable,
        *loglike_args    : Any,
        shape_dict       : dict = {},
        **loglike_kwargs : Any
        ) -> Array:
    """
    Calculates the covariance matrix of the pytree parameters with respect to
    itself. This can be used to optimise the fisher information of the pytree
    model. Simply calcluates the data from the `.model()` method and passes it
    to the `covariance_matrix` function. The `shaped_dict` parameter is used to
    specify the shape of the differentiated vector for specific parameters. For
    example, if the parameter `param` is a 1D array of shape (5,) and we wanted
    to calculate the covariance matrix of the mean, we can pass in
    `shape_dict={'param': (1,)}`. This will differentiate the log likelihood
    with respect to the mean of the parameters.

    Parameters
    ----------
    pytree : Base
        Pytree with a .model() function.
    parameters : Union[str, list]
        A path or list of paths or list of nested paths.
    loglike_fn : callable
        The log likelihood function to differentiate.
    *loglike_args : Any
        The args to pass to the log likelihood function.
    shape_dict : dict = {}
        A dictionary specifying the shape of the differentiated vector for
        specific parameters.
    **loglike_kwargs : Any
        The kwargs to pass to the log likelihood function.
    
    Returns
    -------
    covariance_matrix : Array
        The covariance matrix of the pytree parameters.
    """
    data = pytree.model()
    loglike_args = [pytree.model()] + list(loglike_args)
    return covariance_matrix(pytree, parameters, loglike_fn, *loglike_args, 
        shape_dict=shape_dict, **loglike_kwargs)


def _perturb(
        X          : Array, 
        pytree     : Base(), 
        parameters : Params, 
        shapes     : list, 
        lengths    : list
        ) -> Base():
    """
    Perturbs the pytree parameters by the values in X, automatically setting
    up the correct sizes based on each parameter.

    Parameters
    ----------
    X : Array
        The vector to perturb the parameters by.
    pytree : Base
        Pytree with a .model() function.
    parameters : Union[str, list]
        A path or list of paths or list of nested paths.
    shapes : list
        A list of the shapes of the parameters.
    lengths : list
        A list of the lengths of the parameters.
    
    Returns
    -------
    pytree : Base
        The perturbed pytree.
    """
    # Improve with lax.scan or lax.carry?
    n, xs = 0, []
    if isinstance(parameters, str):
        parameters = [parameters]
    indexes = range(len(parameters))
    for i, param, shape, length in zip(indexes, parameters, shapes, lengths):
        if length == 1:
            xs.append(X[i+n])
        else:
            xs.append(lax.dynamic_slice(X, (i+n,), (length,)).reshape(shape))
            n += length
    return pytree.add(parameters, xs)


# Functions for calculating lengths and shapes for 'dynamically' generated X
# Vectors for the hessian calculation.
def _lengths_to_N(lengths : list, N : int = 0) -> int:
    """
    Converts an list of shapes into a single length.

    Parameters
    ----------
    lengths : list
        A list of the lengths of the parameters.
    N : int = 0
        The current length.
    """
    if len(lengths) == 1:
        return lengths[0] + N
    return _lengths_to_N(lengths[1:], N=lengths[0] + N)


def _shape_to_length(shape : Union[tuple, int], length : int = 0) -> int:
    """
    Converts a shape into a single length.

    Parameters
    ----------
    shape : Union[tuple, int]
        The shape of the parameter.
    length : int = 0
        The current length.

    Returns
    -------
    length : int
        The length of the parameter.
    """
    if isinstance(shape, int):
        return shape
    elif len(shape) == 1:
        return shape[0]
    return _shape_to_length(shape[1:], length=shape[0]*length)


def _get_shape(
        pytree     : Base(), 
        parameter  : str, 
        shape_dict : dict = {}) -> Union[tuple, int]:
    """
    Gets the shape of a parameter in the pytree. If the parameter is in the
    shape_dict, then the shape is taken from there. Otherwise, the shape is
    taken from the pytree.

    Parameters
    ----------
    pytree : Base
        Pytree with a .model() function.
    parameter : str
        The path to the parameter.
    shape_dict : dict = {}
        A dictionary specifying the shape of the differentiated vector for
        specific parameters.
    
    Returns
    -------
    shape : Union[tuple, int]
        The shape of the parameter.
    """
    shape = pytree.get(parameter).shape
    if parameter in shape_dict:
        return shape_dict[parameter]
    return 1 if shape == () else shape


def _shapes_and_lengths(
        pytree     : Base(), 
        parameters : Params, 
        shape_dict : dict = {}) -> tuple:
    """
    Calculates the shapes and lenths of the parameters in the pytree. If the
    parameter is in the shape_dict, then the shape is taken from there.
    Otherwise, the shape is taken from the pytree.

    Parameters
    ----------
    pytree : Base
        Pytree with a .model() function.
    parameters : Union[str, list]
        A path or list of paths or list of nested paths.
    shape_dict : dict = {}
        A dictionary specifying the shape of the differentiated vector for
    
    Returns
    -------
    shapes : list
        A list of the shapes of the parameters.
    """
    if isinstance(parameters, str):
        parameters = [parameters]
    shapes = [_get_shape(pytree, p, shape_dict) for p in parameters]
    lengths = [_shape_to_length(shape) for shape in shapes]
    return shapes, lengths


from __future__ import annotations
from jax import config
config.update("jax_debug_nans", True)
import zodiax

paths = [
    'param',
    'b.param',
    ['param', 'b.param'],
]

def fn(pytree):
    return pytree.param * pytree.b.param


def test_filter_grad(create_base):
    pytree = create_base()
    for path in paths:
        zodiax.eqx.filter_grad(path)(fn)(pytree)


def test_filter_value_and_grad(create_base):
    pytree = create_base()
    for path in paths:
        zodiax.eqx.filter_value_and_grad(path)(fn)(pytree)


def test_partition(create_base):
    pytree = create_base()
    for path in paths:
        zodiax.eqx.partition(pytree, path)



from __future__ import annotations
from jax import config
config.update("jax_debug_nans", True)
from optax import adam, GradientTransformation, MultiTransformState
import zodiax


# Paths
paths = [
    'param',
    'b.param',
    ['param', 'b.param'],
]

# Optimisers
optimisers = [
    adam(0),
    adam(1),
    [adam(0), adam(1)],
]


def test_get_optimiser(create_base):
    """
    tests the get_optimiser method
    """
    # Define parameters and construct base
    pytree = create_base()

    # Test paths
    for path, optimiser in zip(paths, optimisers):
        zodiax.optimisation.get_optimiser(pytree, path, optimiser)


from __future__ import annotations
from jax import config
config.update("jax_debug_nans", True)
import zodiax


# Paths
paths = [
    'param',
    'b.param',
    ['param', 'b.param'],
]


def test_boolean_filter(create_base):
    pytree = create_base()
    for path in paths:
        zodiax.tree.boolean_filter(pytree, path)


def test_set_array(create_base):
    pytree = create_base()
    for path in paths:
        zodiax.tree.set_array(pytree, path)


name = "zodiax"
__version__ = "0.4.1"

# Import as modules
from . import base
from . import tree
from . import optimisation
from . import experimental
from . import eqx
from . import bayes

# Import core functions from modules
from .base import *
from .tree import *
from .optimisation import *
from .eqx import *
from .bayes import *

# Add to __all__
__all__ = base.__all__ + tree.__all__ + optimisation.__all__ + \
    experimental.__all__ + eqx.__all__ + bayes.__all__


from __future__ import annotations
import zodiax
import jax.numpy as np
from equinox import tree_at, Module
from typing import Union, Any, Callable, List
from jaxtyping import Array


__all__ = ["Base"]


Params = Union[str, List[str]]


def _get_leaf(pytree : Base, param : Params) -> Any:
    """
    A hidden class desinged to recurse down a pytree following the param,
    returning the leaf at the end of the param.

    Base case: len(param) == 1
        In this case the leaf referred to by the single param entry is
        returned (and hence recursively sent up to the initial call).

    Recursive case: len(param) > 1
        In this case the function takes the Base like object referred to
        by the first entry in param, and recursively calls this function
        with this new pytree object and the param without the first entry.

    Parameters
    ----------
    pytree : Base
        The pytee object to recurse though.
    param : Params
        The param to recurse down.

    Returns
    -------
    leaf : Any
        The leaf object specified at the end of the param object.
    """
    key = param[0]
    if hasattr(pytree, key):
        pytree = getattr(pytree, key)
    elif isinstance(pytree, dict):
        pytree = pytree[key]
    elif isinstance(pytree, (list, tuple)):
        pytree = pytree[int(key)]
    else:
        raise ValueError(
            "key: {} not found in object: {}".format(key,type(pytree)))

    # Return param if at the end of param, else recurse
    return pytree if len(param) == 1 else _get_leaf(pytree, param[1:])


def _get_leaves(pytree : Base, parameters : list) -> list:
    """
    Returns a list of leaves specified by the parameters.

    Parameters
    ----------
    pytree : Base
        The pytee object to recurse though.
    parameters : list
        A list/tuple of nested parameters. Note param objects can only be
        nested a single time.

    Returns
    -------
    leaves : list
        The list of leaf objects specified by the parameters object
    """
    return [_get_leaf(pytree, param) for param in parameters]


def _unwrap(parameters : Params, values_in : list = None) -> list:
    """
    Unwraps the provided parameters in to the correct list-based format for the
    _get_leaves and _get_leaf methods, returning a single dimensional list
    of input parameters.

    Parameters
    ----------
    parameters : Params
        A list/tuple of nested parameters to unwrap.
    values_in : list = None
        The list of values to be unwrapped.

    Returns
    -------
    parameters, values : list, list
        The list of unwrapped parameters or parameters and values.
    """
    # Inititalise empty lists
    parameters_out, values_out = [], []

    # If values are provided, apply transformation to both
    if values_in is not None:
        # Make sure values is list
        values = values_in if isinstance(values_in, list) else [values_in]

        # Repeat values to match length of parameters
        if len(values) == 1:
            values = values * len(parameters)
        
        # Ensure correct length
        if len(values) != len(parameters):
            raise ValueError(
                "The number of values must match the number of parameters.")

        # Iterate over parameters and values
        for param, value in zip(parameters, values):

            # Recurse and add in the case of list inputs
            if isinstance(param, list):
                new_parameters, new_values = _unwrap(param, value)
                parameters_out  += new_parameters
                values_out += new_values

            # Params must already be absolute
            else:
                parameters_out.append(param)
                values_out.append(value)
        return parameters_out, values_out

    # Just parameters provided
    else:
        # Iterate over parameters
        for param in parameters:

            # Recurse and add in the case of list inputs
            if isinstance(param, list):
                new_parameters = _unwrap(param)
                parameters_out += new_parameters

            # Params must already be absolute
            else:
                parameters_out.append(param)
        return parameters_out


def _format(parameters : Params, values : list = None) -> list:
    """
    Formats the provided parameters in to the correct list-based format for the
    _get_leaves and _get_leaf methods, returning a single dimensional list
    of input parameters.

    Parameters
    ----------
    parameters : Params
        A list/tuple of nested parameters to unwrap.
    values : list = None
        The list of values to be unwrapped.

    Returns
    -------
    parameters, values : list, list
        The list of unwrapped parameters or parameters and values.
    """
    # Nested/multiple inputs
    if isinstance(parameters, list):

        # If there is nesting, ensure correct dis
        if len(parameters) > 1 and values is not None \
            and True in [isinstance(p, list) for p in parameters]:
            assert isinstance(values, list) and len(values) == len(parameters), \
            ("If a list of parameters is provided, the list of values must be "
                "of equal length.")

        # Its a list - iterate and unbind all the keys
        if values is not None:
            flat_parameters, new_values = _unwrap(parameters, values)
        else:
            flat_parameters = _unwrap(parameters)
        
        # Turn into seperate strings
        new_parameters = [param.split('.') if '.' in param else [param] \
                        for param in flat_parameters]

    # Un-nested/singular input
    else:
        # Turn into seperate strings
        new_parameters = [parameters.split('.') if '.' in parameters else [parameters]]
        new_values = [values]

    # Return
    return new_parameters if values is None else (new_parameters, new_values)


###############
### Classes ###
###############
class Base(Module):
    """
    Extend the Equninox.Module class to give a user-friendly 'param based' API
    for working with pytrees by adding a series of methods used to interface
    with the leaves of the pytree using parameters.
    """
    def get(self : Base, parameters : Params) -> Any:
        """
        Get the leaf specified by param.

        Parameters
        ----------
        parameters : Params
            A list/tuple of nested parameters to unwrap.

        Returns
        -------
        leaf, leaves : Any, list
            The leaf or list of leaves specified by parameters.
        """
        new_parameters = _format(parameters)
        values = _get_leaves(self, new_parameters)
        return values[0] if len(new_parameters) == 1 else values


    def set(self       : Base,
            parameters : Params,
            values     : Union[List[Any], Any]) -> Base:
        """
        Set the leaves specified by parameters with values.

        Parameters
        ----------
        parameters : Params
            A param or list of parameters or list of nested parameters.
        values : Union[List[Any], Any]
            The list of values to set at the leaves specified by parameters.

        Returns
        -------
        pytree : Base
            The pytree with leaves specified by parameters updated with values.
        """
        # Allow None inputs
        if values is None:
            values = [None]
            if isinstance(parameters, str):
                parameters = [parameters]
        new_parameters, new_values = _format(parameters, values)

        # Define 'where' function and update pytree
        leaves_fn = lambda pytree: _get_leaves(pytree, new_parameters)
        return tree_at(leaves_fn, self, new_values,
                      is_leaf = lambda leaf: leaf is None)


    def add(self       : Base,
            parameters : Params,
            values     : Union[List[Any], Any]) -> Base:
        """
        Add to the the leaves specified by parameters with values.

        Parameters
        ----------
        parameters : Params
            A param or list of parameters or list of nested parameters.
        values : Union[List[Any], Any]
            The list of values to add to the leaves specified by parameters.

        Returns
        -------
        pytree : Base
            The pytree with values added to leaves specified by parameters.
        """
        new_parameters, new_values = _format(parameters, values)
        new_values = [leaf + value for value, leaf in zip(new_values, \
                                    _get_leaves(self, new_parameters))]

        # Define 'where' function and update pytree
        leaves_fn = lambda pytree: _get_leaves(pytree, new_parameters)
        return tree_at(leaves_fn, self, new_values,
                      is_leaf = lambda leaf: leaf is None)


    def multiply(self       : Base,
                 parameters : Params,
                 values     : Union[List[Any], Any]) -> Base:
        """
        Multiplies the the leaves specified by parameters with values.

        Parameters
        ----------
        parameters : Params
            A param or list of parameters or list of nested parameters.
        values : Union[List[Any], Any]
            The list of values to multiply the leaves specified by parameters.

        Returns
        -------
        pytree : Base
            The pytree with values multiplied by leaves specified by parameters.
        """
        new_parameters, new_values = _format(parameters, values)
        new_values = [leaf * value for value, leaf in zip(new_values, \
                                    _get_leaves(self, new_parameters))]

        # Define 'where' function and update pytree
        leaves_fn = lambda pytree: _get_leaves(pytree, new_parameters)
        return tree_at(leaves_fn, self, new_values,
                      is_leaf = lambda leaf: leaf is None)


    def divide(self       : Base,
               parameters : Params,
               values     : Union[List[Any], Any]) -> Base:
        """
        Divides the the leaves specified by parameters with values.

        Parameters
        ----------
        parameters : Params
            A param or list of parameters or list of nested parameters.
        values : Union[List[Any], Any]
            The list of values to divide the leaves specified by parameters.

        Returns
        -------
        pytree : Base
            The pytree with values divided by leaves specified by parameters.
        """
        new_parameters, new_values = _format(parameters, values)
        new_values = [leaf / value for value, leaf in zip(new_values, \
                                    _get_leaves(self, new_parameters))]

        # Define 'where' function and update pytree
        leaves_fn = lambda pytree: _get_leaves(pytree, new_parameters)
        return tree_at(leaves_fn, self, new_values,
                      is_leaf = lambda leaf: leaf is None)


    def power(self       : Base,
              parameters : Params,
              values     : Union[List[Any], Any]) -> Base:
        """
        Raises th leaves specified by parameters to the power of values.

        Parameters
        ----------
        parameters : Params
            A param or list of parameters or list of nested parameters.
        values : Union[List[Any], Any]
            The list of values to take the leaves specified by parameters to the
            power of.

        Returns
        -------
        pytree : Base
            The pytree with the leaves specified by parameters raised to the power
            of values.
        """
        new_parameters, new_values = _format(parameters, values)
        new_values = [leaf ** value for value, leaf in zip(new_values, \
                                    _get_leaves(self, new_parameters))]

        # Define 'where' function and update pytree
        leaves_fn = lambda pytree: _get_leaves(pytree, new_parameters)
        return tree_at(leaves_fn, self, new_values,
                      is_leaf = lambda leaf: leaf is None)


    def min(self       : Base,
            parameters : Params,
            values     : Union[List[Any], Any]) -> Base:
        """
        Updates the leaves specified by parameters with the minimum value of the
        leaves and values.

        Parameters
        ----------
        parameters : Params
            A param or list of parameters or list of nested parameters.
        values : Union[List[Any], Any]
            The list of values to take the minimum of and the leaf.

        Returns
        -------
        pytree : Base
            The pytree with the leaves specified by parameters updated with the
            minimum value of the leaf and values.
        """
        new_parameters, new_values = _format(parameters, values)
        new_values = [np.minimum(leaf, value) for value, leaf in \
                    zip(new_values, _get_leaves(self, new_parameters))]

        # Define 'where' function and update pytree
        leaves_fn = lambda pytree: _get_leaves(pytree, new_parameters)
        return tree_at(leaves_fn, self, new_values,
                      is_leaf = lambda leaf: leaf is None)


    def max(self       : Base,
            parameters : Params,
            values     : Union[List[Any], Any]) -> Base:
        """
        Updates the leaves specified by parameters with the maximum value of the
        leaves and values.


        Parameters
        ----------
        parameters : Params
            A param or list of parameters or list of nested parameters.
        values : Union[List[Any], Any]
            The list of values to take the maximum of and the leaf.

        Returns
        -------
        pytree : Base
            The pytree with the leaves specified by parameters updated with the
            maximum value of the leaf and values.
        """
        new_parameters, new_values = _format(parameters, values)
        new_values = [np.maximum(leaf, value) for value, leaf in \
                    zip(new_values, _get_leaves(self, new_parameters))]

        # Define 'where' function and update pytree
        leaves_fn = lambda pytree: _get_leaves(pytree, new_parameters)
        return tree_at(leaves_fn, self, new_values,
                      is_leaf = lambda leaf: leaf is None)


"""
Extra methods for possible future classes that recursively searches for 
parameters.
"""
#     Method 1, get first
#     def __getattr__(self, key):
#         """

#         """
#         # Found it, nice work
#         dict_like = self.__dict__
#         if key in dict_like.keys():
#             return dict_like[key]

#         # Expand and iterate though items
#         for value in dict_like.values():

#             # Dictionary, call the recursive method
#             if isinstance(value, dict):
#                 try:
#                     return _recurse_dict(value, key)
#                 except ValueError as e:
#                     pass

#             # dLux object, recurse
#             if isinstance(value, Base):
#                 try:
#                     return getattr(value, key)
#                 except ValueError as e:
#                     pass

#         # Not found, raise error
#         raise ValueError("'{}' object has no attribute '{}'"\
#                              .format(type(self), key))


#     def _recurse_dict(self, dict_like, key):
#         """

#         """
#         # Return item if it exists
#         if key in dict_like:
#             return dict_like[key]

#         # Iterate through values
#         for value in dict_like.values():

#             # Value is a dict, Recurse
#             if isinstance(value, dict):
#                 try:
#                     return self._recurse_dict(value, key)
#                 except ValueError as e:
#                     pass

#             # Value is a dLux object, recall the getattr method
#             if isinstance(value, Base):
#                 try:
#                     return getattr(value, key)
#                 except ValueError as e:
#                     pass

#         # Nothing found, raise Error
#         raise ValueError("'{}' not found.".format(key))


#     # Method 2, get all
#     def __getattr__(self, key):
#         """

#         """
#         return self._get_all(key, [])


#     def _get_all(self, key, values):

#         # Found it, nice work
#         dict_like = self.__dict__
#         if key in dict_like.keys():
#             values.append(dict_like[key])

#         # Expand and iterate though items
#         for value in dict_like.values():

#             # Dictionary, call the recursive method
#             if isinstance(value, dict):
#                 # values.append(self._recurse_dict(value, key, values))
#                 values = self._recurse_dict(value, key, values)

#             # dLux object, recurse
#             if isinstance(value, Base):
#                 # values.append(value._get_all(key, values))
#                 values = value._get_all(key, values)

#         return values


#     def _recurse_dict(self, dict_like, key, values):
#         """

#         """
#         # Return item if it exists
#         if key in dict_like:
#             values.append(dict_like[key])

#         # Iterate through values
#         for value in dict_like.values():

#             # Value is a dict, Recurse
#             if isinstance(value, dict):
#                 # values.append(self._recurse_dict(value, key, values))
#                 values = self._recurse_dict(value, key, values)

#             # Value is a dLux object, recall the getattr method
#             if isinstance(value, Base):
#                 # values.append(value._get_all(key, values))
#                 values = value._get_all(key, values)

#         return values


from __future__ import annotations
from jax import config
config.update("jax_debug_nans", True)
import jax.numpy as np
import zodiax
import pytest


# Paths
paths = [
    'param',
    'b.param',
    ['param', 'b.param'],
]


def test_poiss_loglike(create_base):
    pytree = create_base()
    data = pytree.model()
    zodiax.bayes.poiss_loglike(pytree, data)


def test_chi2_loglike(create_base):
    pytree = create_base()
    data = pytree.model()
    zodiax.bayes.chi2_loglike(pytree, data, noise=2)


def test_fisher_matrix(create_base):
    pytree = create_base()
    data = pytree.model()
    loglike_fn = zodiax.bayes.poiss_loglike
    shape_dict = {'param': (1,)}

    for param in paths:
        zodiax.bayes.fisher_matrix(pytree, param, loglike_fn, data, 
            shape_dict=shape_dict)


def test_covariance_matrix(create_base):
    pytree = create_base()
    data = pytree.model()
    loglike_fn = zodiax.bayes.poiss_loglike
    shape_dict = {'param': (1,)}

    for param in paths:
        zodiax.bayes.covariance_matrix(pytree, param, loglike_fn, data, 
            shape_dict=shape_dict)


def test_covaraince_entropy(create_base):
    pytree = create_base()
    data = pytree.model()
    loglike_fn = zodiax.bayes.poiss_loglike
    shape_dict = {'param': (1,)}
    for param in paths:
        cov = zodiax.bayes.covariance_matrix(pytree, param, loglike_fn, data,
            shape_dict=shape_dict)
        zodiax.bayes.covaraince_entropy(cov)


def test_self_fisher_matrix(create_base):
    pytree = create_base()
    loglike_fn = zodiax.bayes.poiss_loglike
    shape_dict = {'param': (1,)}

    for param in paths:
        zodiax.bayes.self_fisher_matrix(pytree, param, loglike_fn, 
            shape_dict=shape_dict)


def test_self_covariance_matrix(create_base):
    pytree = create_base()
    loglike_fn = zodiax.bayes.poiss_loglike
    shape_dict = {'param': (1,)}

    for param in paths:
        zodiax.bayes.self_covariance_matrix(pytree, param, loglike_fn, 
            shape_dict=shape_dict)


from __future__ import annotations
from jax import config
config.update("jax_debug_nans", True)
import jax.numpy as np
import zodiax
import pytest


def test_unwrap():
    """
    Test the _unwrap method
    """

    # Test unwrapping
    wrapped_a = ['a', ['b', ['c', ['d']]]]
    wrapped_b = [[[['a'], 'b'], 'c'], 'd']
    wrapped_c = ['a', 'b', 'c', 'd']

    assert zodiax.base._unwrap(wrapped_a) == ['a', 'b', 'c', 'd']
    assert zodiax.base._unwrap(wrapped_b) == ['a', 'b', 'c', 'd']
    assert zodiax.base._unwrap(wrapped_c) == ['a', 'b', 'c', 'd']

    # Test with values
    wrapped_a = ['a', ['b', 'c', 'd']]
    wrapped_b = [['a', 'b', 'c'], 'd']
    wrapped_c = ['a', 'b', 'c', 'd']

    assert zodiax.base._unwrap(wrapped_a, [1, 2])[1]       == [1, 2, 2, 2]
    assert zodiax.base._unwrap(wrapped_b, [1, 2])[1]       == [1, 1, 1, 2]
    assert zodiax.base._unwrap(wrapped_c, [1, 2, 3, 4])[1] == [1, 2, 3, 4]


def test_format():
    """
    test the _format method
    """

    # Test formatting
    path_a = ['a.b', ['b.c', ['c.d', ['d.e']]]]
    path_b = [[[['a.b'], 'b.c'], 'c.d'], 'd.e']
    path_c = ['a.b', 'b.c', 'c.d', 'd.e']

    assert zodiax.base._format(path_a) == [['a', 'b'], ['b', 'c'], ['c', 'd'], ['d', 'e']]
    assert zodiax.base._format(path_b) == [['a', 'b'], ['b', 'c'], ['c', 'd'], ['d', 'e']]
    assert zodiax.base._format(path_c) == [['a', 'b'], ['b', 'c'], ['c', 'd'], ['d', 'e']]

    # Test with values
    path_a = ['a.b', ['b.c', 'c.d', 'd.e']]
    path_b = [['a.b', 'b.c', 'c.d'], 'd.e']
    path_c = ['a.b', 'b.c', 'c.d', 'd.e']

    assert zodiax.base._format(path_a, [1, 2])[1]       == [1, 2, 2, 2]
    assert zodiax.base._format(path_b, [1, 2])[1]       == [1, 1, 1, 2]
    assert zodiax.base._format(path_c, [1, 2, 3, 4])[1] == [1, 2, 3, 4]


class TestBase():
    """
    Tests the Base class.
    """

    def test_get(self, create_base):
        """
        tests the get method
        """
        # Test single parameter
        create_base().get('param')

        # Test multiple parameters
        create_base().get(['param', 'b.param'])


    def test_set(self, create_base):
        """
        tests the set method
        """
        # Test single parameter
        create_base().set('param', 10.)

        # Test multiple parameters
        create_base().set(['param', 'b.param'], [10., 10.])


    def test_add(self, create_base):
        """
        tests the add method
        """
        # Test single parameter
        create_base().add('param', 10.)

        # Test multiple parameters
        create_base().add(['param', 'b.param'], [10., 10.])


    def test_multiply(self, create_base):
        """
        tests the multiply method
        """
        # Test single parameter
        create_base().multiply('param', 10.)

        # Test multiple parameters
        create_base().multiply(['param', 'b.param'], [10., 10.])


    def test_divide(self, create_base):
        """
        tests the divide method
        """
        # Test single parameter
        create_base().divide('param', 10.)

        # Test multiple parameters
        create_base().divide(['param', 'b.param'], [10., 10.])


    def test_power(self, create_base):
        """
        tests the power method
        """
        # Test single parameter
        create_base().power('param', 10.)

        # Test multiple parameters
        create_base().power(['param', 'b.param'], [10., 10.])


    def test_min(self, create_base):
        """
        tests the min method
        """
        # Test single parameter
        create_base().min('param', 10.)

        # Test multiple parameters
        create_base().min(['param', 'b.param'], [10., 10.])


    def test_max(self, create_base):
        """
        tests the max method
        """
        # Test single parameter
        create_base().max('param', 10.)

        # Test multiple parameters
        create_base().max(['param', 'b.param'], [10., 10.])


import zodiax
import equinox
from functools import wraps
from jaxtyping import PyTree
from typing import Union, Callable, List
# from equinox import partition, combine
import equinox as eqx
from types import ModuleType


Params = Union[str, List[str]]
Base = lambda: zodiax.base.Base


def filter_grad(
    parameters : Params, 
    *filter_args,
    **filter_kwargs
    ) -> Callable:
    """
    Applies the equinox filter_grad function to the input parameters. The 
    corresponding equinox docs are found [here](https://docs.kidger.site/
    equinox/api/filtering/transformations/)

    Parameters
    ----------
    parameters : Union[str, List[str]]
        The parameters to filter. Can either be a single string path or a list
        of paths.
    *filter_args : Any
        The args to pass to the equinox filter_grad function.
    **filter_kwargs : Any
        The kwargs to pass to the equinox filter_grad function.
    
    Returns
    -------
    Callable
        The wrapped function.
    """
    def wrapper(func : Callable):
        
        @wraps(func)
        def inner_wrapper(pytree : PyTree, *args, **kwargs):

            # Convert parameters
            boolean_filter = zodiax.tree.boolean_filter(pytree, parameters)

            # Wrap original function
            @equinox.filter_grad(*filter_args, **filter_kwargs)
            def recombine(traced : PyTree, static : PyTree):
                return func(eqx.combine(traced, static), *args, **kwargs)
            
            # Return wrapped function
            return recombine(*eqx.partition(pytree, boolean_filter))
        return inner_wrapper
    return wrapper


def filter_value_and_grad(
    parameters : Params, 
    *filter_args, 
    **filter_kwargs
    ) -> Callable:
    """
    Applies the equinox filter_value_and_grad function to the input parameters.
    The corresponding equinox docs are found [here](https://docs.kidger.site/
    equinox/api/filtering/transformations/)

    Parameters
    ----------
    parameters : Union[str, List[str]]
        The parameters to filter. Can either be a single string path or a list
        of paths.
    *filter_args : Any
        The args to pass to the equinox filter_value_and_grad function.
    **filter_kwargs : Any
        The kwargs to pass to the equinox filter_value_and_grad function.
    
    Returns
    -------
    Callable
        The wrapped function.
    """
    def wrapper(func : Callable):

        @wraps(func)
        def inner_wrapper(pytree : PyTree, *args, **kwargs):

            # Convert parameters
            boolean_filter = zodiax.tree.boolean_filter(pytree, parameters)

            # Wrap original function
            @equinox.filter_value_and_grad(*filter_args, **filter_kwargs)
            def recombine(traced : PyTree, static : PyTree):
                return func(eqx.combine(traced, static), *args, **kwargs)
            
            # Return wrapped function
            return recombine(*eqx.partition(pytree, boolean_filter))
        return inner_wrapper
    return wrapper


def partition(
    pytree : Base(), 
    parameters : Params, 
    *partition_args, 
    **partition_kwargs) -> tuple:
    """
    Wraps the equinox partition function to take in a list of parameters to
    partition. The corresponding equinox docs are found [here](https://docs.
    kidger.site/equinox/api/filtering/transformations/)

    Parameters
    ----------
    pytree : Base()
        The pytree to partition.
    parameters : Union[str, List[str]]
        The parameters to partition. Can either be a single string path or a
        list of paths.
    *partition_args : Any
        The args to pass to the equinox partition function.
    **partition_kwargs : Any
        The kwargs to pass to the equinox partition function.
    
    Returns
    -------
    pytree1 : Base()
        A matching pytree with Nones at all leaves not specified by the
        parameters.
    pytree2 : Base()
        A matching pytree with Nones at all leaves specified by the parameters.
    """
    if isinstance(parameters, str):
        parameters = [parameters]
    boolean_filter = zodiax.tree.boolean_filter(pytree, parameters)
    return equinox.partition(pytree, boolean_filter, *partition_args,
        **partition_kwargs)


# Dictionary of replaced functions
replaced_dict = {
    'filter_grad'           : filter_grad,
    'filter_value_and_grad' : filter_value_and_grad,
    'partition'             : partition,
}


# Use the __all__ attribute of the external package to get a list of all 
# public functions
external_functions = [func_name for func_name in dir(equinox) 
    if not func_name.startswith("_")]


# Define what function we want to overwrite
replace = list(replaced_dict.keys())


# Create a dictionary of wrapper functions that simply call the corresponding 
# function from the external package
wrapper_functions = {}
replaced_functions = []
for func_name in external_functions:
    param = getattr(equinox, func_name)

    # Import functions in the namespace and push them to zodiax namespace
    if callable(param):
        replaced_functions.append(func_name)
        if func_name in replaced_dict.keys():
            wrapper_functions[func_name] = replaced_dict[func_name]
        else:
            wrapper_functions[func_name] = getattr(equinox, func_name)
    
    # If it is a module import it as a submodule
    elif isinstance(param, ModuleType):
        wrapper_functions[func_name] = getattr(equinox, func_name)
    
    # The rest of these should just be custom types that we dont need
    else:
        pass


# Export the wrapper functions from the module
__all__ = replaced_functions
globals().update(wrapper_functions)


import zodiax
from jax.tree_util import tree_map
from equinox import is_array, filter as eqx_filter
from optax import adam, multi_transform, GradientTransformation
from typing import Union, List
from jaxtyping import PyTree


__all__ = ["get_optimiser"]


Base = lambda: zodiax.base.Base
Params = Union[str, List[str]]
Optimisers = Union[GradientTransformation, list]


def get_optimiser(pytree     : Base(),
                  parameters : Params,
                  optimisers : Optimisers,
                  ) -> tuple:
    """
    Returns an Optax.GradientTransformion object, with the optimisers
    specified by optimisers applied to the leaves specified by parameters.

    Parameters
    ----------
    pytree : Base
        A zodiax.base.Base object.
    parameters :  Union[str, List[str]]
        A path or list of parameters or list of nested parameters.
    optimisers : Union[optax.GradientTransformation, list]
        A optax.GradientTransformation or list of
        optax.GradientTransformation objects to be applied to the leaves
        specified by parameters.

    Returns
    -------
    optimiser : optax.GradientTransformion
        TODO Update
        A tuple of (Optax.GradientTransformion, optax.MultiTransformState)
        objects, with the optimisers applied to the leaves specified by
        parameters, and the initialised optimisation state.
    state : optax.MultiTransformState
    """
    # Pre-wrap single inputs into a list since optimisers have a length of 2
    if not isinstance(optimisers, list):
        optimisers = [optimisers]
    
    # parameters have to default be wrapped in a list to match optimiser
    if isinstance(parameters, str):
        parameters = [parameters]

    # Construct groups and get param_spec
    groups = [str(i) for i in range(len(optimisers))]
    param_spec = tree_map(lambda _: "null", pytree)
    param_spec = param_spec.set(parameters, groups)

    # Generate optimiser dictionary and Assign the null group
    opt_dict = dict([(groups[i], optimisers[i]) \
                        for i in range(len(groups))])
    opt_dict["null"] = adam(0.0)

    # Get optimiser object and filtered optimiser
    optim = multi_transform(opt_dict, param_spec)
    opt_state = optim.init(eqx_filter(pytree, is_array))

    # Return
    return (optim, opt_state)


import zodiax
import jax.numpy as np
import jax.tree_util as jtu
from jax import Array
from typing import Union, List, Any


__all__ = ['improve_jit_hash']


Base = lambda: zodiax.base.Base
Params = Union[str, List[str]]


# Improved jit pytree
def _float_from_0d(leaf : Any):
    """
    Turns a 0d array into a float for hashing under jit.
    """
    if isinstance(leaf, Array):
        return float(leaf) if leaf.ndim == 0 else leaf
    else:
        return leaf

def improve_jit_hash(pytree : Base(), parameters : Params):
    """
    # TODO: Cast all leaves not specified by parameters to a numpy (not jax)
    array and all 0d arrays to floats, plus any other possile non-jax types. 
    This could improve performance under jit since this apparently changes the
    jaxpr under the hood, and could potentially result in them being treated
    as static values. This is currently not used since it is not clear if it
    actually improves performance.

    This *could* also be set to be the default actions in a zodiax version
    of equinox filter_jit. Would require changing default equinox filter
    function to be jax type arrays only.


    Replaces all scalar array leaves with python floats. This allows for those
    leaves to be marked as static under jit. The arguments in the params input
    are not cast to python types to prevent the pytree from being recompiled
    every time the arguments change under an optimisation loop.

    Parameters
    ----------
    pytree : Base
        The pytree to be jitted.
    parameters : Union[List[str], str]
        The arguments to be optimised under the jit compilation.
    
    Returns
    -------
    pytree : zodiax.Base
        The pytree with scalar arrays leaves replaced with python floats.
    """
    dynamic, static = zodiax.eqx.partition(pytree, parameters)
    static = jtu.tree_map(_float_from_0d, static)
    return zodiax.combine(dynamic, static)


import zodiax
import jax.numpy as np
import jax.tree_util as jtu
from jax import Array
from jaxtyping import PyTree
from typing import Union, List, Any
from equinox import partition, combine


__all__ = ['boolean_filter', 'set_array']


Base = lambda: zodiax.base.Base
Params = Union[str, List[str]]


# Boolean
def boolean_filter(
    pytree     : Base(), 
    parameters : Params, 
    inverse    : bool = False) -> Base():
    """
    Returns a pytree of matching structure with boolean values at the leaves.
    Leaves specified by paths will be True, all others will be False. 

    TODO: Possibly improve by setting both true and false simultaneously.
    Maybe do this with jax keypaths?

    Parameters
    ----------
    pytree : PyTree
        The pytree to be filtered.
    parameters : Union[str, list]
        A path or list of paths or list of nested paths.
    inverse : bool = False
        If True, the boolean values will be inverted, by default False

    Returns
    -------
    args : PyTree
        An pytree of matching structre with boolean values at the leaves.
    """
    parameters = parameters if isinstance(parameters, list) else [parameters]
    if not inverse:
        false_pytree = jtu.tree_map(lambda _: False, pytree)
        return false_pytree.set(parameters, len(parameters) * [True])
    else:
        true_pytree = jtu.tree_map(lambda _: True, pytree)
        return true_pytree.set(parameters, len(parameters) * [False])


# Array
def _to_array(leaf : Any):
    if not isinstance(leaf, Array):
        return np.asarray(leaf, dtype=float)
    else:
        return leaf

def set_array(pytree : Base(), parameters : Params) -> Base():
    """
    Converts all leaves specified by parameters in the pytree to arrays to 
    ensure they have a .shape property for static dimensionality and size
    checks. This allows for 'dynamicly generated' array shapes from the path
    based `parameters` input. This is used for dynamically generating the
    latent X parameter that we need to generate in order to calculate the
    hessian.

    Parameters
    ----------
    pytree : Base()
        The pytree to be converted.
    parameters : Params
        The leaves to be converted to arrays.
    
    Returns
    -------
    pytree : Base()
        The pytree with the specified leaves converted to arrays.
    """
    new_leaves = jtu.tree_map(_to_array, pytree.get(parameters))
    return pytree.set(parameters, new_leaves)